import configparser
import os
import sys
import textwrap
from collections.abc import Callable
from functools import partial
from pathlib import Path
from typing import Any, NoReturn

if sys.version_info[:2] >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

INI_USAGE = """
(config)
...
[mypy.plugins.django-stubs]
django_settings_module = str (default: `os.getenv("DJANGO_SETTINGS_MODULE")`)
strict_settings = bool (default: true)
...
"""
TOML_USAGE = """
(config)
...
[tool.django-stubs]
django_settings_module = str (default: `os.getenv("DJANGO_SETTINGS_MODULE")`)
strict_settings = bool (default: true)
...
"""
INVALID_FILE = "mypy config file is not specified or found"
COULD_NOT_LOAD_FILE = "could not load configuration file"
MISSING_SECTION = "no section [{section}] found"
DJANGO_SETTINGS_ENV_VAR = "DJANGO_SETTINGS_MODULE"
MISSING_DJANGO_SETTINGS = f"missing required 'django_settings_module' config.\
 Either specify this config or set your `{DJANGO_SETTINGS_ENV_VAR}` env var"
INVALID_BOOL_SETTING = "invalid {key!r}: the setting must be a boolean"


def exit_with_error(msg: str, is_toml: bool = False) -> NoReturn:
    """Using mypy's argument parser, raise `SystemExit` to fail hard if validation fails.

    Considering that the plugin's startup duration is around double as long as mypy's, this aims to
    import and construct objects only when that's required - which happens once and terminates the
    run. Considering that most of the runs are successful, there's no need for this to linger in the
    global scope.
    """
    from mypy.main import CapturableArgumentParser

    handler = CapturableArgumentParser(
        prog="(django-stubs) mypy", usage=textwrap.dedent(TOML_USAGE if is_toml else INI_USAGE)
    )
    handler.error(msg)


class DjangoPluginConfig:
    __slots__ = ("django_settings_module", "strict_settings")
    django_settings_module: str
    strict_settings: bool

    def __init__(self, config_file: str | None) -> None:
        if not config_file:
            exit_with_error(INVALID_FILE)

        filepath = Path(config_file)
        if not filepath.is_file():
            exit_with_error(INVALID_FILE)

        if filepath.suffix.lower() == ".toml":
            self.parse_toml_file(filepath)
        else:
            self.parse_ini_file(filepath)

    def parse_toml_file(self, filepath: Path) -> None:
        toml_exit: Callable[[str], NoReturn] = partial(exit_with_error, is_toml=True)
        try:
            with filepath.open(mode="rb") as f:
                data = tomllib.load(f)
        except (tomllib.TOMLDecodeError, OSError):
            toml_exit(COULD_NOT_LOAD_FILE)

        try:
            config: dict[str, Any] = data["tool"]["django-stubs"]
        except KeyError:
            toml_exit(MISSING_SECTION.format(section="tool.django-stubs"))

        django_settings_module = config.get("django_settings_module") or os.getenv(DJANGO_SETTINGS_ENV_VAR)
        if not django_settings_module:
            toml_exit(MISSING_DJANGO_SETTINGS)

        self.django_settings_module = django_settings_module

        if not isinstance(self.django_settings_module, str):
            toml_exit("invalid 'django_settings_module': the setting must be a string")

        self.strict_settings = config.get("strict_settings", True)
        if not isinstance(self.strict_settings, bool):
            toml_exit(INVALID_BOOL_SETTING.format(key="strict_settings"))

    def parse_ini_file(self, filepath: Path) -> None:
        parser = configparser.ConfigParser()
        try:
            with filepath.open(encoding="utf-8") as f:
                parser.read_file(f, source=str(filepath))
        except OSError:
            exit_with_error(COULD_NOT_LOAD_FILE)

        section = "mypy.plugins.django-stubs"
        if not parser.has_section(section):
            exit_with_error(MISSING_SECTION.format(section=section))

        if parser.has_option(section, "django_settings_module"):
            django_settings_module = parser.get(section, "django_settings_module").strip("'\"")
        else:
            django_settings_module = os.getenv(DJANGO_SETTINGS_ENV_VAR, "")

        if not django_settings_module:
            exit_with_error(MISSING_DJANGO_SETTINGS)

        self.django_settings_module = django_settings_module

        try:
            self.strict_settings = parser.getboolean(section, "strict_settings", fallback=True)
        except ValueError:
            exit_with_error(INVALID_BOOL_SETTING.format(key="strict_settings"))

    def to_json(self, extra_data: dict[str, Any]) -> dict[str, Any]:
        """We use this method to reset mypy cache via `report_config_data` hook."""
        return {
            "django_settings_module": self.django_settings_module,
            "strict_settings": self.strict_settings,
            **dict(sorted(extra_data.items())),
        }



from typing import Any, ClassVar

from django.db import models
from django.db.models.expressions import Func

class CumeDist(Func):
    output_field: ClassVar[models.FloatField]

class DenseRank(Func):
    output_field: ClassVar[models.IntegerField]

class FirstValue(Func): ...

class LagLeadFunction(Func):
    def __init__(self, expression: str | None, offset: int = 1, default: int | None = None, **extra: Any) -> None: ...

class Lag(LagLeadFunction): ...
class LastValue(Func): ...
class Lead(LagLeadFunction): ...

class NthValue(Func):
    def __init__(self, expression: str | None, nth: int = 1, **extra: Any) -> None: ...

class Ntile(Func):
    def __init__(self, num_buckets: int = 1, **extra: Any) -> None: ...
    output_field: ClassVar[models.IntegerField]

class PercentRank(Func):
    output_field: ClassVar[models.FloatField]

class Rank(Func):
    output_field: ClassVar[models.IntegerField]

class RowNumber(Func):
    output_field: ClassVar[models.IntegerField]

__all__ = [
    "CumeDist",
    "DenseRank",
    "FirstValue",
    "Lag",
    "LastValue",
    "Lead",
    "NthValue",
    "Ntile",
    "PercentRank",
    "Rank",
    "RowNumber",
]



from ipaddress import IPv6Address
from typing import Any

MAX_IPV6_ADDRESS_LENGTH: int

def clean_ipv6_address(
    ip_str: Any, unpack_ipv4: bool = False, error_message: str = ..., max_length: int = 39
) -> str: ...
def is_valid_ipv6_address(ip_addr: str | IPv6Address) -> bool: ...



from django.contrib.gis.db.models.sql.conversion import AreaField as AreaField
from django.contrib.gis.db.models.sql.conversion import DistanceField as DistanceField

__all__ = ["AreaField", "DistanceField"]



from collections.abc import Iterable
from typing import Any

from django.template.base import Origin, Template
from django.template.engine import Engine

class Loader:
    engine: Engine
    get_template_cache: dict[str, Any]
    def __init__(self, engine: Engine) -> None: ...
    def get_template(self, template_name: str, skip: list[Origin] | None = None) -> Template: ...
    def get_template_sources(self, template_name: str) -> Iterable[Origin]: ...
    def reset(self) -> None: ...



from typing import TypeVar

from django.db import models

_T = TypeVar("_T", bound=models.Model)

class CurrentSiteManager(models.Manager[_T]):
    def __init__(self, field_name: str | None = None) -> None: ...



from typing import Any

from django.contrib.messages.storage.base import BaseStorage
from django.http.request import HttpRequest
from django.utils.functional import _StrOrPromise

class MessageFailure(Exception): ...

def add_message(
    request: HttpRequest | None,
    level: int,
    message: _StrOrPromise,
    extra_tags: str = "",
    fail_silently: bool | str = False,
) -> None: ...
def get_messages(request: HttpRequest) -> list[Any] | BaseStorage: ...
def get_level(request: HttpRequest) -> int: ...
def set_level(request: HttpRequest, level: int) -> bool: ...
def debug(
    request: HttpRequest,
    message: _StrOrPromise,
    extra_tags: str = "",
    fail_silently: bool | str = False,
) -> None: ...
def info(
    request: HttpRequest,
    message: _StrOrPromise,
    extra_tags: str = "",
    fail_silently: bool | str = False,
) -> None: ...
def success(
    request: HttpRequest,
    message: _StrOrPromise,
    extra_tags: str = "",
    fail_silently: bool | str = False,
) -> None: ...
def warning(
    request: HttpRequest,
    message: _StrOrPromise,
    extra_tags: str = "",
    fail_silently: bool | str = False,
) -> None: ...
def error(
    request: HttpRequest,
    message: _StrOrPromise,
    extra_tags: str = "",
    fail_silently: bool | str = False,
) -> None: ...

__all__ = (
    "add_message",
    "get_messages",
    "get_level",
    "set_level",
    "debug",
    "info",
    "success",
    "warning",
    "error",
    "MessageFailure",
)



from collections.abc import Iterable
from decimal import Decimal

def format(
    number: Decimal | float | str,
    decimal_sep: str,
    decimal_pos: int | None = None,
    grouping: int | Iterable[int] = 0,
    thousand_sep: str = "",
    force_grouping: bool = False,
    use_l10n: bool | None = None,
) -> str: ...



from collections import OrderedDict
from collections.abc import Sequence

from django.core.exceptions import FieldError
from django.db.models.base import Model
from django.db.models.fields.related import RelatedField
from django.db.models.fields.reverse_related import ForeignObjectRel
from mypy.checker import TypeChecker
from mypy.nodes import ARG_NAMED, ARG_NAMED_OPT, Expression
from mypy.plugin import FunctionContext, MethodContext
from mypy.types import AnyType, Instance, TupleType, TypedDictType, TypeOfAny, get_proper_type
from mypy.types import Type as MypyType
from mypy.typevars import fill_typevars

from mypy_django_plugin.django.context import DjangoContext, LookupsAreUnsupported
from mypy_django_plugin.lib import fullnames, helpers
from mypy_django_plugin.lib.helpers import parse_bool
from mypy_django_plugin.transformers.models import get_annotated_type


def _extract_model_type_from_queryset(queryset_type: Instance, api: TypeChecker) -> Instance | None:
    if queryset_type.type.has_base(fullnames.MANAGER_CLASS_FULLNAME):
        to_model_fullname = helpers.get_manager_to_model(queryset_type.type)
        if to_model_fullname is not None:
            to_model = helpers.lookup_fully_qualified_typeinfo(api, to_model_fullname)
            if to_model is not None:
                to_model_instance = fill_typevars(to_model)
                assert isinstance(to_model_instance, Instance)
                return to_model_instance

    for base_type in [queryset_type, *queryset_type.type.bases]:
        if not len(base_type.args):
            continue
        model = get_proper_type(base_type.args[0])
        if isinstance(model, Instance) and helpers.is_model_type(model.type):
            return model
    return None


def determine_proper_manager_type(ctx: FunctionContext) -> MypyType:
    default_return_type = get_proper_type(ctx.default_return_type)
    assert isinstance(default_return_type, Instance)

    outer_model_info = helpers.get_typechecker_api(ctx).scope.active_class()
    if (
        outer_model_info is None
        or not helpers.is_model_type(outer_model_info)
        or outer_model_info.self_type is None
        or not default_return_type.type.is_generic()
    ):
        return default_return_type

    return helpers.reparametrize_instance(default_return_type, [outer_model_info.self_type])


def get_field_type_from_lookup(
    ctx: MethodContext,
    django_context: DjangoContext,
    model_cls: type[Model],
    *,
    method: str,
    lookup: str,
    silent_on_error: bool = False,
) -> MypyType | None:
    try:
        lookup_field, model_cls = django_context.resolve_lookup_into_field(model_cls, lookup)
    except FieldError as exc:
        if not silent_on_error:
            ctx.api.fail(exc.args[0], ctx.context)
        return None
    except LookupsAreUnsupported:
        return AnyType(TypeOfAny.explicit)

    if lookup_field is None:
        return AnyType(TypeOfAny.implementation_artifact)
    elif (isinstance(lookup_field, RelatedField) and lookup_field.column == lookup) or isinstance(
        lookup_field, ForeignObjectRel
    ):
        model_cls = django_context.get_field_related_model_cls(lookup_field)
        lookup_field = django_context.get_primary_key_field(model_cls)

    api = helpers.get_typechecker_api(ctx)
    model_info = helpers.lookup_class_typeinfo(api, model_cls)
    field_get_type = django_context.get_field_get_type(api, model_info, lookup_field, method=method)
    return field_get_type


def get_values_list_row_type(
    ctx: MethodContext,
    django_context: DjangoContext,
    model_cls: type[Model],
    *,
    is_annotated: bool,
    flat: bool,
    named: bool,
) -> MypyType:
    field_lookups = resolve_field_lookups(ctx.args[0], django_context)
    if field_lookups is None:
        return AnyType(TypeOfAny.from_error)

    typechecker_api = helpers.get_typechecker_api(ctx)
    model_info = helpers.lookup_class_typeinfo(typechecker_api, model_cls)
    if len(field_lookups) == 0:
        if flat:
            primary_key_field = django_context.get_primary_key_field(model_cls)
            lookup_type = get_field_type_from_lookup(
                ctx, django_context, model_cls, lookup=primary_key_field.attname, method="values_list"
            )
            assert lookup_type is not None
            return lookup_type
        elif named:
            column_types: OrderedDict[str, MypyType] = OrderedDict()
            for field in django_context.get_model_fields(model_cls):
                column_type = django_context.get_field_get_type(
                    typechecker_api, model_info, field, method="values_list"
                )
                column_types[field.attname] = column_type
            if is_annotated:
                # Return a NamedTuple with a fallback so that it's possible to access any field
                return helpers.make_oneoff_named_tuple(
                    typechecker_api,
                    "Row",
                    column_types,
                    extra_bases=[typechecker_api.named_generic_type(fullnames.ANY_ATTR_ALLOWED_CLASS_FULLNAME, [])],
                )
            else:
                return helpers.make_oneoff_named_tuple(typechecker_api, "Row", column_types)
        else:
            # flat=False, named=False, all fields
            if is_annotated:
                return typechecker_api.named_generic_type("builtins.tuple", [AnyType(TypeOfAny.special_form)])
            field_lookups = []
            for field in django_context.get_model_fields(model_cls):
                field_lookups.append(field.attname)

    if len(field_lookups) > 1 and flat:
        typechecker_api.fail("'flat' is not valid when 'values_list' is called with more than one field", ctx.context)
        return AnyType(TypeOfAny.from_error)

    column_types = OrderedDict()
    for field_lookup in field_lookups:
        lookup_field_type = get_field_type_from_lookup(
            ctx, django_context, model_cls, lookup=field_lookup, method="values_list", silent_on_error=is_annotated
        )
        if lookup_field_type is None:
            if is_annotated:
                lookup_field_type = AnyType(TypeOfAny.from_omitted_generics)
            else:
                return AnyType(TypeOfAny.from_error)
        column_types[field_lookup] = lookup_field_type

    if flat:
        assert len(column_types) == 1
        row_type = next(iter(column_types.values()))
    elif named:
        row_type = helpers.make_oneoff_named_tuple(typechecker_api, "Row", column_types)
    else:
        # Since there may have been repeated field lookups, we cannot just use column_types.values here.
        # This is not the case in named above, because Django will error if duplicate fields are requested.
        resolved_column_types = [column_types[field_lookup] for field_lookup in field_lookups]
        row_type = helpers.make_tuple(typechecker_api, resolved_column_types)

    return row_type


def extract_proper_type_queryset_values_list(ctx: MethodContext, django_context: DjangoContext) -> MypyType:
    # called on the Instance, returns QuerySet of something
    if not isinstance(ctx.type, Instance):
        return ctx.default_return_type
    default_return_type = get_proper_type(ctx.default_return_type)
    if not isinstance(default_return_type, Instance):
        return ctx.default_return_type

    model_type = _extract_model_type_from_queryset(ctx.type, helpers.get_typechecker_api(ctx))
    if model_type is None:
        return AnyType(TypeOfAny.from_omitted_generics)

    is_annotated = helpers.is_annotated_model(model_type.type)
    model_cls = (
        django_context.get_model_class_by_fullname(model_type.type.bases[0].type.fullname)
        if is_annotated
        else django_context.get_model_class_by_fullname(model_type.type.fullname)
    )
    if model_cls is None:
        return default_return_type

    flat_expr = helpers.get_call_argument_by_name(ctx, "flat")
    if flat_expr is not None:
        flat = parse_bool(flat_expr)
    else:
        flat = False

    named_expr = helpers.get_call_argument_by_name(ctx, "named")
    if named_expr is not None:
        named = parse_bool(named_expr)
    else:
        named = False

    if flat and named:
        ctx.api.fail("'flat' and 'named' can't be used together", ctx.context)
        return helpers.reparametrize_instance(default_return_type, [model_type, AnyType(TypeOfAny.from_error)])

    # account for possible None
    flat = flat or False
    named = named or False

    row_type = get_values_list_row_type(
        ctx, django_context, model_cls, is_annotated=is_annotated, flat=flat, named=named
    )
    return helpers.reparametrize_instance(default_return_type, [model_type, row_type])


def gather_kwargs(ctx: MethodContext) -> dict[str, MypyType] | None:
    num_args = len(ctx.arg_kinds)
    kwargs = {}
    named = (ARG_NAMED, ARG_NAMED_OPT)
    for i in range(num_args):
        if not ctx.arg_kinds[i]:
            continue
        if any(kind not in named for kind in ctx.arg_kinds[i]):
            # Only named arguments supported
            return None
        for j in range(len(ctx.arg_names[i])):
            name = ctx.arg_names[i][j]
            assert name is not None
            kwargs[name] = ctx.arg_types[i][j]
    return kwargs


def extract_proper_type_queryset_annotate(ctx: MethodContext, django_context: DjangoContext) -> MypyType:
    # called on the Instance, returns QuerySet of something
    if not isinstance(ctx.type, Instance):
        return ctx.default_return_type
    default_return_type = get_proper_type(ctx.default_return_type)
    if not isinstance(default_return_type, Instance):
        return ctx.default_return_type

    model_type = _extract_model_type_from_queryset(ctx.type, helpers.get_typechecker_api(ctx))
    if model_type is None:
        return AnyType(TypeOfAny.from_omitted_generics)

    api = helpers.get_typechecker_api(ctx)

    field_types: dict[str, MypyType] | None = None
    kwargs = gather_kwargs(ctx)
    if kwargs:
        # For now, we don't try to resolve the output_field of the field would be, but use Any.
        # NOTE: It's important that we use 'special_form' for 'Any' as otherwise we can
        # get stuck with mypy interpreting an overload ambiguity towards the
        # overloaded 'Field.__get__' method when its 'model' argument gets matched. This
        # is because the model argument gets matched with a model subclass that is
        # parametrized with a type that contains the 'Any' below and then mypy runs in
        # to a (false?) ambiguity, due to 'Any', and can't decide what overload/return
        # type to select
        #
        # Example:
        #   class MyModel(models.Model):
        #       field = models.CharField()
        #
        #   # Our plugin auto generates the following subclass
        #   class MyModel_WithAnnotations(MyModel, django_stubs_ext.Annotations[_Annotations]):
        #       ...
        #   # Assume
        #   x = MyModel.objects.annotate(foo=F("id")).get()
        #   reveal_type(x)  # MyModel_WithAnnotations[TypedDict({"foo": Any})]
        #   # Then, on an attribute access of 'field' like
        #   reveal_type(x.field)
        #
        # Now 'CharField.__get__', which is overloaded, is passed 'x' as the 'model'
        # argument and mypy consider it ambiguous to decide which overload method to
        # select due to the 'Any' in 'TypedDict({"foo": Any})'. But if we specify the
        # 'Any' as 'TypeOfAny.special_form' mypy doesn't consider the model instance to
        # contain 'Any' and the ambiguity goes away.
        field_types = {name: AnyType(TypeOfAny.special_form) for name, _ in kwargs.items()}

    fields_dict = None
    if field_types is not None:
        fields_dict = helpers.make_typeddict(
            api,
            fields=field_types,
            required_keys=set(field_types.keys()),
            readonly_keys=set(),
        )

    if fields_dict is not None:
        annotated_type = get_annotated_type(api, model_type, fields_dict=fields_dict)
    else:
        annotated_type = model_type

    row_type: MypyType
    if len(default_return_type.args) > 1:
        original_row_type = get_proper_type(default_return_type.args[1])
        row_type = original_row_type
        if isinstance(original_row_type, TypedDictType):
            row_type = api.named_generic_type(
                "builtins.dict", [api.named_generic_type("builtins.str", []), AnyType(TypeOfAny.from_omitted_generics)]
            )
        elif isinstance(original_row_type, TupleType):
            fallback: Instance = original_row_type.partial_fallback
            if fallback is not None and fallback.type.has_base("typing.NamedTuple"):
                # TODO: Use a NamedTuple which contains the known fields, but also
                #  falls back to allowing any attribute access.
                row_type = AnyType(TypeOfAny.implementation_artifact)
            else:
                row_type = api.named_generic_type("builtins.tuple", [AnyType(TypeOfAny.from_omitted_generics)])
        elif isinstance(original_row_type, Instance) and helpers.is_model_type(original_row_type.type):
            row_type = annotated_type
    else:
        row_type = annotated_type
    return helpers.reparametrize_instance(default_return_type, [annotated_type, row_type])


def resolve_field_lookups(lookup_exprs: Sequence[Expression], django_context: DjangoContext) -> list[str] | None:
    field_lookups = []
    for field_lookup_expr in lookup_exprs:
        field_lookup = helpers.resolve_string_attribute_value(field_lookup_expr, django_context)
        if field_lookup is None:
            return None
        field_lookups.append(field_lookup)
    return field_lookups


def extract_proper_type_queryset_values(ctx: MethodContext, django_context: DjangoContext) -> MypyType:
    # called on QuerySet, return QuerySet of something
    if not isinstance(ctx.type, Instance):
        return ctx.default_return_type
    default_return_type = get_proper_type(ctx.default_return_type)
    if not isinstance(default_return_type, Instance):
        return ctx.default_return_type

    model_type = _extract_model_type_from_queryset(ctx.type, helpers.get_typechecker_api(ctx))
    if model_type is None:
        return AnyType(TypeOfAny.from_omitted_generics)

    model_cls = django_context.get_model_class_by_fullname(model_type.type.fullname)
    if model_cls is None:
        return default_return_type

    field_lookups = resolve_field_lookups(ctx.args[0], django_context)
    if field_lookups is None:
        return AnyType(TypeOfAny.from_error)

    if len(field_lookups) == 0:
        for field in django_context.get_model_fields(model_cls):
            field_lookups.append(field.attname)

    column_types: OrderedDict[str, MypyType] = OrderedDict()
    for field_lookup in field_lookups:
        field_lookup_type = get_field_type_from_lookup(
            ctx, django_context, model_cls, lookup=field_lookup, method="values"
        )
        if field_lookup_type is None:
            return helpers.reparametrize_instance(default_return_type, [model_type, AnyType(TypeOfAny.from_error)])

        column_types[field_lookup] = field_lookup_type

    row_type = helpers.make_typeddict(ctx.api, column_types, set(column_types.keys()), set())
    return helpers.reparametrize_instance(default_return_type, [model_type, row_type])



from collections.abc import Sequence
from email.message import Message
from email.mime.base import MIMEBase
from email.mime.message import MIMEMessage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import make_msgid as make_msgid
from typing import Any, NamedTuple, TypeAlias, overload

from django.utils.functional import _StrOrPromise

utf8_charset: Any
utf8_charset_qp: Any
DEFAULT_ATTACHMENT_MIME_TYPE: str
RFC5322_EMAIL_LINE_LENGTH_LIMIT: int

class BadHeaderError(ValueError): ...

ADDRESS_HEADERS: set[str]

def forbid_multi_line_headers(name: str, val: str, encoding: str) -> tuple[str, str]: ...
def sanitize_address(addr: tuple[str, str] | str, encoding: str) -> str: ...

class MIMEMixin:
    def as_string(self, unixfrom: bool = False, linesep: str = "\n") -> str: ...
    def as_bytes(self, unixfrom: bool = False, linesep: str = "\n") -> bytes: ...

class SafeMIMEMessage(MIMEMixin, MIMEMessage): ...  # type: ignore[misc]

class SafeMIMEText(MIMEMixin, MIMEText):  # type: ignore[misc]
    encoding: str
    def __init__(self, _text: str, _subtype: str = "plain", _charset: str | None = None) -> None: ...

class SafeMIMEMultipart(MIMEMixin, MIMEMultipart):  # type: ignore[misc]
    encoding: str
    def __init__(
        self,
        _subtype: str = "mixed",
        boundary: Any | None = None,
        _subparts: Any | None = None,
        encoding: str | None = None,
        **_params: Any,
    ) -> None: ...

_AttachmentContent: TypeAlias = bytes | EmailMessage | Message | SafeMIMEText | str
_AttachmentTuple: TypeAlias = (
    tuple[str, _AttachmentContent] | tuple[str | None, _AttachmentContent, str] | tuple[str, _AttachmentContent, None]
)

class EmailAlternative(NamedTuple):
    content: _AttachmentContent
    mimetype: str

class EmailAttachment(NamedTuple):
    filename: str | None
    content: _AttachmentContent
    mimetype: str

class EmailMessage:
    content_subtype: str
    mixed_subtype: str
    encoding: Any
    to: list[str]
    cc: list[Any]
    bcc: list[Any]
    reply_to: list[Any]
    from_email: str
    subject: _StrOrPromise
    body: _StrOrPromise
    attachments: list[Any]
    extra_headers: dict[Any, Any]
    connection: Any
    def __init__(
        self,
        subject: _StrOrPromise = "",
        body: _StrOrPromise | None = "",
        from_email: str | None = None,
        to: Sequence[str] | None = None,
        bcc: Sequence[str] | None = None,
        connection: Any | None = None,
        attachments: Sequence[MIMEBase | _AttachmentTuple] | None = None,
        headers: dict[str, str] | None = None,
        cc: Sequence[str] | None = None,
        reply_to: Sequence[str] | None = None,
    ) -> None: ...
    def get_connection(self, fail_silently: bool = False) -> Any: ...
    # TODO: when typeshed gets more types for email.Message, move it to MIMEMessage, now it has too many false-positives
    def message(self) -> Any: ...
    def recipients(self) -> list[str]: ...
    def send(self, fail_silently: bool = False) -> int: ...
    @overload
    def attach(self, filename: MIMEBase | None = None, content: None = None, mimetype: None = None) -> None: ...
    @overload
    def attach(
        self, filename: None = None, content: _AttachmentContent | None = None, mimetype: str | None = None
    ) -> None: ...
    @overload
    def attach(
        self, filename: str | None = None, content: _AttachmentContent | None = None, mimetype: str | None = None
    ) -> None: ...
    def attach_file(self, path: str, mimetype: str | None = None) -> None: ...

class EmailMultiAlternatives(EmailMessage):
    alternative_subtype: str
    alternatives: list[tuple[_AttachmentContent, str]]
    def __init__(
        self,
        subject: _StrOrPromise = "",
        body: _StrOrPromise | None = "",
        from_email: str | None = None,
        to: Sequence[str] | None = None,
        bcc: Sequence[str] | None = None,
        connection: Any | None = None,
        attachments: Sequence[MIMEBase | _AttachmentTuple] | None = None,
        headers: dict[str, str] | None = None,
        alternatives: list[tuple[_AttachmentContent, str]] | None = None,
        cc: Sequence[str] | None = None,
        reply_to: Sequence[str] | None = None,
    ) -> None: ...
    def attach_alternative(self, content: _AttachmentContent, mimetype: str) -> None: ...
    def body_contains(self, text: str) -> bool: ...



from typing import Any

from typing_extensions import deprecated

@deprecated(
    "django.utils.itercompat.is_iterable() is deprecated and will be removed in Django 6.0. Use isinstance(..., collections.abc.Iterable) instead."
)
def is_iterable(x: Any) -> bool: ...



from typing import Any

from django.apps.config import AppConfig
from django.apps.registry import Apps

def create_default_site(
    app_config: AppConfig,
    verbosity: int = 2,
    interactive: bool = True,
    using: str = "default",
    apps: Apps = ...,
    **kwargs: Any,
) -> None: ...



from typing import Any

from django.template import Node
from django.template.base import FilterExpression, NodeList, Parser, Token
from django.template.context import Context

register: Any

class GetAvailableLanguagesNode(Node):
    variable: str
    def __init__(self, variable: str) -> None: ...
    def render(self, context: Context) -> str: ...

class GetLanguageInfoNode(Node):
    lang_code: FilterExpression
    variable: str
    def __init__(self, lang_code: FilterExpression, variable: str) -> None: ...
    def render(self, context: Context) -> str: ...

class GetLanguageInfoListNode(Node):
    languages: FilterExpression
    variable: str
    def __init__(self, languages: FilterExpression, variable: str) -> None: ...
    def get_language_info(self, language: Any) -> Any: ...
    def render(self, context: Context) -> str: ...

class GetCurrentLanguageNode(Node):
    variable: str
    def __init__(self, variable: str) -> None: ...
    def render(self, context: Context) -> str: ...

class GetCurrentLanguageBidiNode(Node):
    variable: str
    def __init__(self, variable: str) -> None: ...
    def render(self, context: Context) -> str: ...

class TranslateNode(Node):
    noop: bool
    asvar: str | None
    message_context: FilterExpression | None
    filter_expression: FilterExpression
    def __init__(
        self,
        filter_expression: FilterExpression,
        noop: bool,
        asvar: str | None = None,
        message_context: FilterExpression | None = None,
    ) -> None: ...
    def render(self, context: Context) -> str: ...

class BlockTranslateNode(Node):
    extra_context: dict[str, FilterExpression]
    singular: list[Token]
    plural: list[Token]
    countervar: str | None
    counter: FilterExpression | None
    message_context: FilterExpression | None
    trimmed: bool
    asvar: str | None
    def __init__(
        self,
        extra_context: dict[str, FilterExpression],
        singular: list[Token],
        plural: list[Token] | None = None,
        countervar: str | None = None,
        counter: FilterExpression | None = None,
        message_context: FilterExpression | None = None,
        trimmed: bool = False,
        asvar: str | None = None,
        tag_name: str = "blocktranslate",
    ) -> None: ...
    def render_token_list(self, tokens: list[Token]) -> tuple[str, list[str]]: ...
    def render(self, context: Context, nested: bool = False) -> str: ...

class LanguageNode(Node):
    nodelist: NodeList
    language: FilterExpression
    def __init__(self, nodelist: NodeList, language: FilterExpression) -> None: ...
    def render(self, context: Context) -> str: ...

def do_get_available_languages(parser: Parser, token: Token) -> GetAvailableLanguagesNode: ...
def do_get_language_info(parser: Parser, token: Token) -> GetLanguageInfoNode: ...
def do_get_language_info_list(parser: Parser, token: Token) -> GetLanguageInfoListNode: ...
def language_name(lang_code: str) -> str: ...
def language_name_translated(lang_code: str) -> str: ...
def language_name_local(lang_code: str) -> str: ...
def language_bidi(lang_code: str) -> bool: ...
def do_get_current_language(parser: Parser, token: Token) -> GetCurrentLanguageNode: ...
def do_get_current_language_bidi(parser: Parser, token: Token) -> GetCurrentLanguageBidiNode: ...
def do_translate(parser: Parser, token: Token) -> TranslateNode: ...
def do_block_translate(parser: Parser, token: Token) -> BlockTranslateNode: ...
def language(parser: Parser, token: Token) -> LanguageNode: ...



from typing import Any, ClassVar

from django.db.models.expressions import Func
from django.db.models.fields import IntegerField
from django.db.models.functions.mixins import FixDurationInputMixin, NumericOutputFieldMixin

class Aggregate(Func):
    filter_template: str
    filter: Any
    allow_distinct: bool
    empty_result_set_value: int | None
    def __init__(self, *expressions: Any, distinct: bool = False, filter: Any | None = None, **extra: Any) -> None: ...

class Avg(FixDurationInputMixin, NumericOutputFieldMixin, Aggregate): ...

class Count(Aggregate):
    output_field: ClassVar[IntegerField]

class Max(Aggregate): ...
class Min(Aggregate): ...
class StdDev(NumericOutputFieldMixin, Aggregate): ...
class Sum(FixDurationInputMixin, Aggregate): ...
class Variance(NumericOutputFieldMixin, Aggregate): ...

__all__ = ["Aggregate", "Avg", "Count", "Max", "Min", "StdDev", "Sum", "Variance"]



from collections.abc import Callable, Iterable, Iterator, Mapping, Sequence
from enum import Enum
from logging import Logger
from re import Pattern
from typing import Any, cast

from django.template.context import Context as Context  # Django: imported for backwards compatibility
from django.template.engine import Engine
from django.template.library import Library
from django.template.loaders.base import Loader
from django.utils.safestring import SafeString

FILTER_SEPARATOR: str
FILTER_ARGUMENT_SEPARATOR: str
VARIABLE_ATTRIBUTE_SEPARATOR: str
BLOCK_TAG_START: str
BLOCK_TAG_END: str
VARIABLE_TAG_START: str
VARIABLE_TAG_END: str
COMMENT_TAG_START: str
COMMENT_TAG_END: str
SINGLE_BRACE_START: str
SINGLE_BRACE_END: str
UNKNOWN_SOURCE: str
tag_re: Pattern[str]
logger: Logger

class TokenType(Enum):
    TEXT = cast(int, ...)
    VAR = cast(int, ...)
    BLOCK = cast(int, ...)
    COMMENT = cast(int, ...)

class VariableDoesNotExist(Exception):
    msg: str
    params: tuple[dict[str, str] | str]
    def __init__(self, msg: str, params: tuple[dict[str, str] | str] = ...) -> None: ...

class Origin:
    name: str
    template_name: bytes | str | None
    loader: Loader | None
    def __init__(self, name: str, template_name: bytes | str | None = None, loader: Loader | None = None) -> None: ...
    @property
    def loader_name(self) -> str | None: ...

class Template:
    name: str | None
    origin: Origin
    engine: Engine
    source: str
    nodelist: NodeList
    def __init__(
        self,
        template_string: Template | str,
        origin: Origin | None = None,
        name: str | None = None,
        engine: Engine | None = None,
    ) -> None: ...
    def render(self, context: Context) -> SafeString: ...
    def compile_nodelist(self) -> NodeList: ...
    def get_exception_info(self, exception: Exception, token: Token) -> dict[str, Any]: ...

def linebreak_iter(template_source: str) -> Iterator[int]: ...

class Token:
    contents: str
    token_type: TokenType
    lineno: int | None
    position: tuple[int, int] | None
    def __init__(
        self,
        token_type: TokenType,
        contents: str,
        position: tuple[int, int] | None = None,
        lineno: int | None = None,
    ) -> None: ...
    def split_contents(self) -> list[str]: ...

class Lexer:
    template_string: str
    verbatim: bool | str
    def __init__(self, template_string: str) -> None: ...
    def tokenize(self) -> list[Token]: ...
    def create_token(self, token_string: str, position: tuple[int, int] | None, lineno: int, in_tag: bool) -> Token: ...

class DebugLexer(Lexer):
    template_string: str
    verbatim: bool | str
    def tokenize(self) -> list[Token]: ...

class Parser:
    tokens: list[Token] | str
    tags: dict[str, Callable]
    filters: dict[str, Callable]
    command_stack: list[tuple[str, Token]]
    libraries: dict[str, Library]
    origin: Origin | None
    def __init__(
        self,
        tokens: list[Token] | str,
        libraries: dict[str, Library] | None = None,
        builtins: list[Library] | None = None,
        origin: Origin | None = None,
    ) -> None: ...
    def parse(self, parse_until: Iterable[str] | None = None) -> NodeList: ...
    def skip_past(self, endtag: str) -> None: ...
    def extend_nodelist(self, nodelist: NodeList, node: Node, token: Token) -> None: ...
    def error(self, token: Token, e: Exception | str) -> Exception: ...
    def invalid_block_tag(self, token: Token, command: str, parse_until: Iterable[str] | None = None) -> Any: ...
    def unclosed_block_tag(self, parse_until: Iterable[str]) -> Any: ...
    def next_token(self) -> Token: ...
    def prepend_token(self, token: Token) -> None: ...
    def delete_first_token(self) -> None: ...
    def add_library(self, lib: Library) -> None: ...
    def compile_filter(self, token: str) -> FilterExpression: ...
    def find_filter(self, filter_name: str) -> Callable: ...

constant_string: str
filter_raw_string: str
filter_re: Pattern[str]

class FilterExpression:
    token: str
    filters: list[Any]
    var: Any
    def __init__(self, token: str, parser: Parser) -> None: ...
    def resolve(self, context: Context, ignore_failures: bool = False) -> Any: ...
    @staticmethod
    def args_check(name: str, func: Callable, provided: list[tuple[bool, Any]]) -> bool: ...

class Variable:
    var: dict[Any, Any] | str
    literal: SafeString | float | None
    lookups: tuple[str, ...] | None
    translate: bool
    message_context: str | None
    def __init__(self, var: dict[Any, Any] | str) -> None: ...
    def resolve(self, context: Mapping[str, Mapping[str, Any]] | Context | int | str) -> Any: ...

class Node:
    must_be_first: bool
    child_nodelists: Any
    origin: Origin
    token: Token | None
    def render(self, context: Context) -> str: ...
    def render_annotated(self, context: Context) -> int | str: ...
    def __iter__(self) -> Iterator[Node]: ...
    def get_nodes_by_type(self, nodetype: type[Node]) -> list[Node]: ...

class NodeList(list[Node]):
    contains_nontext: bool
    def render(self, context: Context) -> SafeString: ...
    def get_nodes_by_type(self, nodetype: type[Node]) -> list[Node]: ...

class TextNode(Node):
    s: str
    def __init__(self, s: str) -> None: ...

def render_value_in_context(value: Any, context: Context) -> str: ...

class VariableNode(Node):
    filter_expression: FilterExpression
    def __init__(self, filter_expression: FilterExpression) -> None: ...

kwarg_re: Pattern[str]

def token_kwargs(bits: Sequence[str], parser: Parser, support_legacy: bool = False) -> dict[str, FilterExpression]: ...



from mypy.plugin import MethodContext
from mypy.types import Type as MypyType
from mypy.types import UninhabitedType, get_proper_type

from mypy_django_plugin.django.context import DjangoContext


def check_querydict_is_mutable(ctx: MethodContext, django_context: DjangoContext) -> MypyType:
    ret_type = get_proper_type(ctx.default_return_type)
    if isinstance(ret_type, UninhabitedType):
        ctx.api.fail("This QueryDict is immutable.", ctx.context)
    return ctx.default_return_type



from django.db.models.fields import CharField, EmailField, TextField

class CICharField(CharField): ...
class CIEmailField(EmailField): ...
class CITextField(TextField): ...

__all__ = ["CICharField", "CIEmailField", "CITextField"]



import json
from typing import Any, TypeAlias

import _typeshed
from django.core.serializers.python import Deserializer as PythonDeserializer
from django.core.serializers.python import Serializer as PythonSerializer

_DeserializerInput: TypeAlias = _typeshed.SupportsRead[bytes | str] | bytes | str

class Serializer(PythonSerializer):
    json_kwargs: dict[str, Any]

class Deserializer(PythonDeserializer):
    def __init__(self, stream_or_string: _DeserializerInput, **options: Any) -> None: ...

class DjangoJSONEncoder(json.JSONEncoder):
    allow_nan: bool
    check_circular: bool
    ensure_ascii: bool
    indent: int
    skipkeys: bool
    sort_keys: bool



from collections.abc import Callable, Iterable, Iterator
from contextlib import AbstractContextManager
from types import TracebackType
from typing import Any, TypeAlias

from django.http.request import HttpRequest
from django.template.base import Node, Origin, Template
from django.template.loader_tags import IncludeNode
from typing_extensions import Self

_ContextKeys: TypeAlias = int | str | Node

_ContextValues: TypeAlias = dict[str, Any] | Context

class ContextPopException(Exception): ...

class ContextDict(dict):
    context: BaseContext
    def __init__(self, context: BaseContext, *args: Any, **kwargs: Any) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None: ...

class BaseContext(Iterable[Any]):
    def __init__(self, dict_: Any | None = None) -> None: ...
    def __copy__(self) -> Self: ...
    def __iter__(self) -> Iterator[Any]: ...
    def push(self, *args: Any, **kwargs: Any) -> ContextDict: ...
    def pop(self) -> ContextDict: ...
    def __setitem__(self, key: _ContextKeys, value: Any) -> None: ...
    def set_upward(self, key: _ContextKeys, value: int | str) -> None: ...
    def __getitem__(self, key: _ContextKeys) -> Any: ...
    def __delitem__(self, key: _ContextKeys) -> None: ...
    def __contains__(self, key: _ContextKeys) -> bool: ...
    def get(self, key: _ContextKeys, otherwise: Any | None = None) -> Any | None: ...
    def setdefault(self, key: _ContextKeys, default: list[Origin] | int | None = None) -> list[Origin] | int | None: ...
    def new(self, values: _ContextValues | None = None) -> Context: ...
    def flatten(self) -> dict[_ContextKeys, dict[_ContextKeys, type[Any] | str] | int | str | None]: ...

class Context(BaseContext):
    dicts: Any
    autoescape: bool
    use_l10n: bool | None
    use_tz: bool | None
    template_name: str | None
    render_context: RenderContext
    template: Template | None
    def __init__(
        self,
        dict_: Any | None = None,
        autoescape: bool = True,
        use_l10n: bool | None = None,
        use_tz: bool | None = None,
    ) -> None: ...
    def bind_template(self, template: Template) -> AbstractContextManager[None]: ...
    def update(self, other_dict: dict[str, Any] | Context) -> ContextDict: ...

class RenderContext(BaseContext):
    dicts: list[dict[IncludeNode | str, str]]
    template: Template | None
    def push_state(self, template: Template, isolated_context: bool = True) -> AbstractContextManager[None]: ...

class RequestContext(Context):
    autoescape: bool
    dicts: list[dict[str, str]]
    render_context: RenderContext
    template_name: str | None
    use_l10n: bool | None
    use_tz: bool | None
    request: HttpRequest
    def __init__(
        self,
        request: HttpRequest,
        dict_: dict[str, Any] | None = None,
        processors: list[Callable] | None = None,
        use_l10n: bool | None = None,
        use_tz: bool | None = None,
        autoescape: bool = True,
    ) -> None: ...
    template: Template | None
    def bind_template(self, template: Template) -> AbstractContextManager[None]: ...
    def new(self, values: _ContextValues | None = None) -> RequestContext: ...

def make_context(context: dict[str, Any] | None, request: HttpRequest | None = None, **kwargs: Any) -> Context: ...



ABSTRACT_BASE_USER_MODEL_FULLNAME = "django.contrib.auth.base_user.AbstractBaseUser"
ABSTRACT_USER_MODEL_FULLNAME = "django.contrib.auth.models.AbstractUser"
PERMISSION_MIXIN_CLASS_FULLNAME = "django.contrib.auth.models.PermissionsMixin"
MODEL_METACLASS_FULLNAME = "django.db.models.base.ModelBase"
MODEL_CLASS_FULLNAME = "django.db.models.base.Model"
FIELD_FULLNAME = "django.db.models.fields.Field"
ARRAY_FIELD_FULLNAME = "django.contrib.postgres.fields.array.ArrayField"
FOREIGN_OBJECT_FULLNAME = "django.db.models.fields.related.ForeignObject"
FOREIGN_KEY_FULLNAME = "django.db.models.fields.related.ForeignKey"
ONETOONE_FIELD_FULLNAME = "django.db.models.fields.related.OneToOneField"
MANYTOMANY_FIELD_FULLNAME = "django.db.models.fields.related.ManyToManyField"
DUMMY_SETTINGS_BASE_CLASS = "django.conf._DjangoConfLazyObject"
AUTH_USER_MODEL_FULLNAME = "django.conf.settings.AUTH_USER_MODEL"

QUERYSET_CLASS_FULLNAME = "django.db.models.query.QuerySet"
BASE_MANAGER_CLASS_FULLNAME = "django.db.models.manager.BaseManager"
MANAGER_CLASS_FULLNAME = "django.db.models.manager.Manager"
RELATED_MANAGER_CLASS = "django.db.models.fields.related_descriptors.RelatedManager"

CHOICES_CLASS_FULLNAME = "django.db.models.enums.Choices"
CHOICES_TYPE_METACLASS_FULLNAME = "django.db.models.enums.ChoicesType"

WITH_ANNOTATIONS_FULLNAME = "django_stubs_ext.annotations.WithAnnotations"
ANNOTATIONS_FULLNAME = "django_stubs_ext.annotations.Annotations"

BASEFORM_CLASS_FULLNAME = "django.forms.forms.BaseForm"
FORM_CLASS_FULLNAME = "django.forms.forms.Form"
MODELFORM_CLASS_FULLNAME = "django.forms.models.ModelForm"

FORM_MIXIN_CLASS_FULLNAME = "django.views.generic.edit.FormMixin"

REVERSE_ONE_TO_ONE_DESCRIPTOR = "django.db.models.fields.related_descriptors.ReverseOneToOneDescriptor"
REVERSE_MANY_TO_ONE_DESCRIPTOR = "django.db.models.fields.related_descriptors.ReverseManyToOneDescriptor"
MANY_TO_MANY_DESCRIPTOR = "django.db.models.fields.related_descriptors.ManyToManyDescriptor"
MANY_RELATED_MANAGER = "django.db.models.fields.related_descriptors.ManyRelatedManager"
RELATED_FIELDS_CLASSES = frozenset(
    (
        FOREIGN_OBJECT_FULLNAME,
        FOREIGN_KEY_FULLNAME,
        ONETOONE_FIELD_FULLNAME,
    )
)

OPTIONS_CLASS_FULLNAME = "django.db.models.options.Options"
QUERYDICT_CLASS_FULLNAME = "django.http.request.QueryDict"

COMBINABLE_EXPRESSION_FULLNAME = "django.db.models.expressions.Combinable"
F_EXPRESSION_FULLNAME = "django.db.models.expressions.F"

ANY_ATTR_ALLOWED_CLASS_FULLNAME = "django_stubs_ext.AnyAttrAllowed"

STR_PROMISE_FULLNAME = "django.utils.functional._StrPromise"

OBJECT_DOES_NOT_EXIST = "django.core.exceptions.ObjectDoesNotExist"
MULTIPLE_OBJECTS_RETURNED = "django.core.exceptions.MultipleObjectsReturned"

DJANGO_ABSTRACT_MODELS = frozenset(
    (
        ABSTRACT_BASE_USER_MODEL_FULLNAME,
        ABSTRACT_USER_MODEL_FULLNAME,
        PERMISSION_MIXIN_CLASS_FULLNAME,
        "django.contrib.sessions.base_session.AbstractBaseSession",
    )
)



from mypy.nodes import TypeInfo
from mypy.plugin import AnalyzeTypeContext
from mypy.semanal import SemanticAnalyzer
from mypy.typeanal import TypeAnalyser
from mypy.types import PlaceholderType, ProperType
from mypy.types import Type as MypyType
from mypy.typevars import fill_typevars_with_any

from mypy_django_plugin.django.context import DjangoContext
from mypy_django_plugin.lib import fullnames, helpers


def get_user_model(ctx: AnalyzeTypeContext, django_context: DjangoContext) -> MypyType:
    assert isinstance(ctx.api, TypeAnalyser)
    assert isinstance(ctx.api.api, SemanticAnalyzer)

    def get_abstract_base_user(api: SemanticAnalyzer) -> ProperType:
        sym = api.lookup_fully_qualified(fullnames.ABSTRACT_BASE_USER_MODEL_FULLNAME)
        assert isinstance(sym.node, TypeInfo)
        return fill_typevars_with_any(sym.node)

    if not django_context.is_contrib_auth_installed:
        return get_abstract_base_user(ctx.api.api)

    auth_user_model = django_context.settings.AUTH_USER_MODEL
    model_info = helpers.resolve_lazy_reference(
        auth_user_model, api=ctx.api.api, django_context=django_context, ctx=ctx.context
    )
    if model_info is None:
        fullname = django_context.model_class_fullnames_by_label.get(auth_user_model)
        if fullname is not None:
            # When we've tried to resolve 'AUTH_USER_MODEL' but got no class back but
            # we notice that its value is recognised we'll return a placeholder for
            # the class as we expect it to exist later on.
            return PlaceholderType(fullname=fullname, args=[], line=ctx.context.line)
        return get_abstract_base_user(ctx.api.api)

    return fill_typevars_with_any(model_info)



from mypy.plugin import MethodContext
from mypy.types import Instance, get_proper_type
from mypy.types import Type as MypyType

from mypy_django_plugin.lib import fullnames, helpers


def get_model_of_related_manager(ctx: MethodContext) -> Instance | None:
    """
    Returns the type parameter (_To) instance of a `RelatedManager` instance when it's a
    model. Otherwise `None` is returned.

    For example: if given a `RelatedManager[A]` where `A` is a model then `A` is
    returned.
    """
    default_return_type = get_proper_type(ctx.default_return_type)
    if (
        isinstance(default_return_type, Instance)
        and default_return_type.type.fullname == fullnames.RELATED_MANAGER_CLASS
    ):
        # This is a call to '__get__' overload with a model instance of
        # 'ReverseManyToOneDescriptor'. Returning a 'RelatedManager'. Which we want to,
        # just like Django, build from the default manager of the related model.
        related_manager = default_return_type
        # Require first type argument of 'RelatedManager' to be a model
        if (
            len(related_manager.args) >= 1
            and isinstance((_To := get_proper_type(related_manager.args[0])), Instance)
            and helpers.is_model_type(_To.type)
        ):
            return _To

    return None


def refine_many_to_one_related_manager(ctx: MethodContext) -> MypyType:
    """
    Updates the 'RelatedManager' returned by e.g. 'ReverseManyToOneDescriptor' to be a subclass
    of 'RelatedManager' and the related model's default manager.
    """
    to_model_instance = get_model_of_related_manager(ctx)
    if to_model_instance is None:
        return ctx.default_return_type

    checker = helpers.get_typechecker_api(ctx)
    related_manager_info = helpers.get_reverse_manager_info(
        checker, to_model_instance.type, derived_from="_default_manager"
    )
    if related_manager_info is None:
        return ctx.default_return_type
    return Instance(related_manager_info, [])



from typing import Any

from django.contrib.auth.models import Permission, _AnyUser, _User, _UserModel, _UserType
from django.db.models import QuerySet
from django.db.models.base import Model
from django.http.request import HttpRequest

UserModel = _UserModel

class BaseBackend:
    def authenticate(self, request: HttpRequest | None, **kwargs: Any) -> _User | None: ...
    async def aauthenticate(self, request: HttpRequest | None, **kwargs: Any) -> _User | None: ...
    def get_user(self, user_id: Any) -> _User | None: ...
    async def aget_user(self, user_id: Any) -> _User | None: ...
    def get_user_permissions(self, user_obj: _AnyUser, obj: Model | None = ...) -> set[str]: ...
    async def aget_user_permissions(self, user_obj: _AnyUser, obj: Model | None = ...) -> set[str]: ...
    def get_group_permissions(self, user_obj: _AnyUser, obj: Model | None = ...) -> set[str]: ...
    async def aget_group_permissions(self, user_obj: _AnyUser, obj: Model | None = ...) -> set[str]: ...
    def get_all_permissions(self, user_obj: _AnyUser, obj: Model | None = ...) -> set[str]: ...
    async def aget_all_permissions(self, user_obj: _AnyUser, obj: Model | None = ...) -> set[str]: ...
    def has_perm(self, user_obj: _AnyUser, perm: str, obj: Model | None = ...) -> bool: ...
    async def ahas_perm(self, user_obj: _AnyUser, perm: str, obj: Model | None = ...) -> bool: ...

class ModelBackend(BaseBackend):
    def authenticate(
        self, request: HttpRequest | None, username: str | None = ..., password: str | None = ..., **kwargs: Any
    ) -> _User | None: ...
    async def aauthenticate(
        self, request: HttpRequest | None, username: str | None = ..., password: str | None = ..., **kwargs: Any
    ) -> _User | None: ...
    def has_module_perms(self, user_obj: _AnyUser, app_label: str) -> bool: ...
    async def ahas_module_perms(self, user_obj: _AnyUser, app_label: str) -> bool: ...
    def user_can_authenticate(self, user: _AnyUser | None) -> bool: ...
    def with_perm(
        self,
        perm: str | Permission,
        is_active: bool = ...,
        include_superusers: bool = ...,
        obj: Model | None = ...,
    ) -> QuerySet[_User]: ...

class AllowAllUsersModelBackend(ModelBackend): ...

class RemoteUserBackend(ModelBackend):
    create_unknown_user: bool
    def authenticate(self, request: HttpRequest | None, remote_user: str) -> _User | None: ...  # type: ignore[override]
    async def aauthenticate(self, request: HttpRequest | None, remote_user: str) -> _User | None: ...  # type: ignore[override]
    def clean_username(self, username: str) -> str: ...
    def configure_user(self, request: HttpRequest, user: _UserType, created: bool = ...) -> _UserType: ...
    async def aconfigure_user(self, request: HttpRequest, user: _UserType, created: bool = ...) -> _UserType: ...

class AllowAllUsersRemoteUserBackend(RemoteUserBackend): ...



from django.db.models import Func
from django.db.models.expressions import Combinable
from django.db.models.fields import Field

class Cast(Func):
    def __init__(self, expression: Combinable | str, output_field: str | Field) -> None: ...

class Coalesce(Func): ...

class Collate(Func):
    def __init__(self, expression: Combinable | str, collation: str) -> None: ...

class Greatest(Func): ...
class Least(Func): ...
class NullIf(Func): ...



from mypy.nodes import MemberExpr, NameExpr, TypeInfo
from mypy.plugin import AttributeContext
from mypy.typeanal import make_optional_type
from mypy.types import AnyType, Instance, TupleType, TypeOfAny, get_proper_type
from mypy.types import Type as MypyType

from mypy_django_plugin.lib import fullnames, helpers


def transform_into_proper_attr_type(ctx: AttributeContext) -> MypyType:
    """
    A `get_attribute_hook` to make `.choices` and `.values` optional if `__empty__` is defined.

    When the `__empty__` label is specified, the `choices` and `values` properties can return a
    blank choice/value. This hook will amend the type returned by those properties.
    """
    if isinstance(ctx.context, MemberExpr):
        method_name = ctx.context.name
    else:
        ctx.api.fail("Unable to resolve type of choices property", ctx.context)
        return AnyType(TypeOfAny.from_error)

    expr = ctx.context.expr

    if isinstance(expr, MemberExpr):
        expr = expr.expr

    if isinstance(expr, NameExpr) and isinstance(expr.node, TypeInfo):
        node = expr.node
    else:
        ctx.api.fail("Unable to resolve type of choices property", ctx.context)
        return AnyType(TypeOfAny.from_error)

    default_attr_type = get_proper_type(ctx.default_attr_type)

    if not node.is_enum or not node.has_base(fullnames.CHOICES_CLASS_FULLNAME):
        return default_attr_type

    # Enums with more than one base will treat the first base as the mixed-in type.
    base_type = node.bases[0] if len(node.bases) > 1 else None

    # When `__empty__` is defined, the `.choices` and `.values` properties will include `None` for
    # the blank choice which is labelled by the value of `__empty__`.
    has_blank_choice = node.get("__empty__") is not None

    if (
        method_name == "choices"
        and isinstance(default_attr_type, Instance)
        and default_attr_type.type.fullname == "builtins.list"
        and len(default_attr_type.args) == 1
    ):
        choice_arg = get_proper_type(default_attr_type.args[0])

        if isinstance(choice_arg, TupleType) and choice_arg.length() == 2:
            value_arg, label_arg = choice_arg.items
            value_arg = get_proper_type(value_arg)

            if isinstance(value_arg, AnyType) and base_type is not None:
                new_value_arg = make_optional_type(base_type) if has_blank_choice else base_type
                new_choice_arg = choice_arg.copy_modified(items=[new_value_arg, label_arg])
                return helpers.reparametrize_instance(default_attr_type, [new_choice_arg])

            elif has_blank_choice:
                new_value_arg = make_optional_type(value_arg)
                new_choice_arg = choice_arg.copy_modified(items=[new_value_arg, label_arg])
                return helpers.reparametrize_instance(default_attr_type, [new_choice_arg])

    elif (
        method_name == "values"
        and isinstance(default_attr_type, Instance)
        and default_attr_type.type.fullname == "builtins.list"
        and len(default_attr_type.args) == 1
    ):
        value_arg = get_proper_type(default_attr_type.args[0])

        if isinstance(value_arg, AnyType) and base_type is not None:
            new_value_arg = make_optional_type(base_type) if has_blank_choice else base_type
            return helpers.reparametrize_instance(default_attr_type, [new_value_arg])

        elif has_blank_choice:
            new_value_arg = make_optional_type(value_arg)
            return helpers.reparametrize_instance(default_attr_type, [new_value_arg])

    elif method_name == "value" and isinstance(default_attr_type, AnyType) and base_type is not None:
        return base_type

    return default_attr_type



from collections.abc import Callable, Sequence
from contextlib import AbstractContextManager
from types import TracebackType
from typing import Any, TypeVar

from django.test import LiveServerTestCase
from typing_extensions import Self

_F = TypeVar("_F", bound=Callable[..., Any])

class SeleniumTestCaseBase:
    browsers: Any
    browser: Any
    @classmethod
    def import_webdriver(cls, browser: Any) -> type[Any]: ...  # Type[WebDriver]
    def create_webdriver(self) -> Any: ...  # WebDriver

class ChangeWindowSize:
    def __init__(self, width: int, height: int, selenium: Any) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None: ...

class SeleniumTestCase(LiveServerTestCase):
    implicit_wait: int
    screenshots: bool = False
    selenium: Any
    def desktop_size(self) -> AbstractContextManager[None]: ...
    def small_screen_size(self) -> AbstractContextManager[None]: ...
    def mobile_size(self) -> AbstractContextManager[None]: ...
    def rtl(self) -> AbstractContextManager[None]: ...
    def dark(self) -> AbstractContextManager[None]: ...
    def set_emulated_media(self, *, media: str | None = None, features: list[dict[str, str]] | None = None) -> None: ...
    def high_contrast(self) -> AbstractContextManager[None]: ...
    def take_screenshot(self, name: str) -> None: ...
    def disable_implicit_wait(self) -> AbstractContextManager[None]: ...

def screenshot_cases(method_names: str | Sequence[str]) -> Callable[[_F], _F]: ...



import os
import tempfile
import uuid
from collections.abc import Generator
from contextlib import contextmanager
from typing import Any
from unittest import mock

import pytest

from mypy_django_plugin.config import DjangoPluginConfig

TEMPLATE = """
(config)
...
[mypy.plugins.django-stubs]
django_settings_module = str (default: `os.getenv("DJANGO_SETTINGS_MODULE")`)
strict_settings = bool (default: true)
...
(django-stubs) mypy: error: {}
"""

TEMPLATE_TOML = """
(config)
...
[tool.django-stubs]
django_settings_module = str (default: `os.getenv("DJANGO_SETTINGS_MODULE")`)
strict_settings = bool (default: true)
...
(django-stubs) mypy: error: {}
"""


@contextmanager
def write_to_file(file_contents: str, suffix: str | None = None) -> Generator[str, None, None]:
    with tempfile.NamedTemporaryFile(mode="w+", suffix=suffix) as config_file:
        config_file.write(file_contents)
        config_file.seek(0)
        yield config_file.name


@pytest.mark.parametrize(
    ("config_file_contents", "message_part"),
    [
        pytest.param(
            ["[not-really-django-stubs]"],
            "no section [mypy.plugins.django-stubs] found",
            id="missing-section",
        ),
        pytest.param(
            ["[mypy.plugins.django-stubs]", "\tnot_django_not_settings_module = badbadmodule"],
            "missing required 'django_settings_module' config.\
 Either specify this config or set your `DJANGO_SETTINGS_MODULE` env var",
            id="missing-settings-module",
        ),
        pytest.param(
            ["[mypy.plugins.django-stubs]"],
            "missing required 'django_settings_module' config.\
 Either specify this config or set your `DJANGO_SETTINGS_MODULE` env var",
            id="no-settings-given",
        ),
        pytest.param(
            ["[mypy.plugins.django-stubs]", "django_settings_module = some.module", "strict_settings = bad"],
            "invalid 'strict_settings': the setting must be a boolean",
            id="missing-settings-module",
        ),
    ],
)
def test_misconfiguration_handling(capsys: Any, config_file_contents: list[str], message_part: str) -> None:
    """Invalid configuration raises `SystemExit` with a precise error message."""
    contents = "\n".join(config_file_contents).expandtabs(4)
    with write_to_file(contents) as filename:
        with pytest.raises(SystemExit, match="2"):
            DjangoPluginConfig(filename)

    error_message = "usage: " + TEMPLATE.format(message_part)
    assert error_message == capsys.readouterr().err


@pytest.mark.parametrize(
    "filename",
    [
        pytest.param(uuid.uuid4().hex, id="not matching an existing file"),
        pytest.param("", id="as empty string"),
        pytest.param(None, id="as none"),
    ],
)
def test_handles_filename(capsys: Any, filename: str) -> None:
    with pytest.raises(SystemExit, match="2"):
        DjangoPluginConfig(filename)

    error_message = "usage: " + TEMPLATE.format("mypy config file is not specified or found")
    assert error_message == capsys.readouterr().err


@pytest.mark.parametrize(
    ("config_file_contents", "message_part"),
    [
        pytest.param(
            """
            [tool.django-stubs]
            django_settings_module = 123
            """,
            "invalid 'django_settings_module': the setting must be a string",
            id="django_settings_module not string",
        ),
        pytest.param(
            """
            [tool.not-really-django-stubs]
            django_settings_module = "my.module"
            """,
            "no section [tool.django-stubs] found",
            id="missing django-stubs section",
        ),
        pytest.param(
            """
            [tool.django-stubs]
            not_django_not_settings_module = "badbadmodule"
            """,
            "missing required 'django_settings_module' config.\
 Either specify this config or set your `DJANGO_SETTINGS_MODULE` env var",
            id="missing django_settings_module",
        ),
        pytest.param(
            "tool.django-stubs]",
            "could not load configuration file",
            id="invalid toml",
        ),
        pytest.param(
            """
            [tool.django-stubs]
            django_settings_module = "some.module"
            strict_settings = "a"
            """,
            "invalid 'strict_settings': the setting must be a boolean",
            id="invalid strict_settings type",
        ),
    ],
)
def test_toml_misconfiguration_handling(capsys: Any, config_file_contents, message_part) -> None:
    with write_to_file(config_file_contents, suffix=".toml") as filename:
        with pytest.raises(SystemExit, match="2"):
            DjangoPluginConfig(filename)

    error_message = "usage: " + TEMPLATE_TOML.format(message_part)
    assert error_message == capsys.readouterr().err


@pytest.mark.parametrize("boolean_value", ["true", "false"])
def test_correct_toml_configuration(boolean_value: str) -> None:
    config_file_contents = f"""
    [tool.django-stubs]
    some_other_setting = "setting"
    django_settings_module = "my.module"
    strict_settings = {boolean_value}
    """

    with write_to_file(config_file_contents, suffix=".toml") as filename:
        config = DjangoPluginConfig(filename)

    assert config.django_settings_module == "my.module"
    assert config.strict_settings is (boolean_value == "true")


@pytest.mark.parametrize("boolean_value", ["true", "True", "false", "False"])
def test_correct_configuration(boolean_value) -> None:
    """Django settings module gets extracted given valid configuration."""
    config_file_contents = "\n".join(
        [
            "[mypy.plugins.django-stubs]",
            "some_other_setting = setting",
            "django_settings_module = my.module",
            f"strict_settings = {boolean_value}",
        ]
    )
    with write_to_file(config_file_contents) as filename:
        config = DjangoPluginConfig(filename)

    assert config.django_settings_module == "my.module"
    assert config.strict_settings is (boolean_value.lower() == "true")


@pytest.mark.parametrize("boolean_value", ["true", "false"])
def test_correct_toml_configuration_with_django_setting_from_env(boolean_value: str) -> None:
    config_file_contents = f"""
    [tool.django-stubs]
    some_other_setting = "setting"
    strict_settings = {boolean_value}
    """
    django_settings_env_value = "my.module"

    with write_to_file(config_file_contents, suffix=".toml") as filename:
        with mock.patch.dict(os.environ, {"DJANGO_SETTINGS_MODULE": django_settings_env_value}):
            config = DjangoPluginConfig(filename)

    assert config.django_settings_module == django_settings_env_value
    assert config.strict_settings is (boolean_value == "true")


@pytest.mark.parametrize("boolean_value", ["true", "True", "false", "False"])
def test_correct_configuration_with_django_setting_from_env(boolean_value) -> None:
    """Django settings module gets extracted given valid configuration."""
    config_file_contents = "\n".join(
        [
            "[mypy.plugins.django-stubs]",
            "some_other_setting = setting",
            f"strict_settings = {boolean_value}",
        ]
    )
    django_settings_env_value = "my.module"

    with write_to_file(config_file_contents) as filename:
        with mock.patch.dict(os.environ, {"DJANGO_SETTINGS_MODULE": django_settings_env_value}):
            config = DjangoPluginConfig(filename)

    assert config.django_settings_module == django_settings_env_value
    assert config.strict_settings is (boolean_value.lower() == "true")
